package conv

import (
	"git.perx.ru/perxis/perxis-go/pkg/invitations"
	"git.perx.ru/perxis/perxis-go/pkg/options"
	pb "git.perx.ru/perxis/perxis-go/proto/invitations"
	timestamp "google.golang.org/protobuf/types/known/timestamppb"
)

type Encoder struct {
}

var _ invitations.ProtoEncoder = &Encoder{}

func (e *Encoder) EncodeAcceptRequest(invitationId string, userId string) *pb.AcceptRequest {
	return &pb.AcceptRequest{
		InvitationId: invitationId,
		UserId:       userId,
	}
}

func (e *Encoder) DecodeAcceptRequest(request *pb.AcceptRequest) (invitationId string, userId string) {
	if request == nil {
		return
	}
	return request.InvitationId, request.UserId
}

func (e *Encoder) EncodeCreateRequest(invitation *invitations.Invitation) *pb.CreateRequest {
	return &pb.CreateRequest{
		Invitation: InvitationToProto(invitation),
	}
}

func (e *Encoder) DecodeCreateRequest(request *pb.CreateRequest) (invitation *invitations.Invitation) {
	if request == nil {
		return
	}
	return InvitationFromProto(request.Invitation)
}

func (e *Encoder) EncodeCreateResponse(response *pb.CreateResponse) (created *invitations.Invitation) {
	if response == nil {
		return
	}
	return InvitationFromProto(response.Invitation)
}

func (e *Encoder) DecodeCreateResponse(created *invitations.Invitation) *pb.CreateResponse {
	return &pb.CreateResponse{
		Invitation: InvitationToProto(created),
	}
}

func (e *Encoder) EncodeDeleteRequest(invitationId string) *pb.DeleteRequest {
	return &pb.DeleteRequest{
		InvitationId: invitationId,
	}
}

func (e *Encoder) DecodeDeleteRequest(request *pb.DeleteRequest) (invitationId string) {
	return request.InvitationId
}

func (e *Encoder) EncodeFindRequest(filter *invitations.Filter, opts *options.FindOptions) *pb.FindRequest {
	request := &pb.FindRequest{}

	if request.Filter != nil {
		filter = &invitations.Filter{
			ID:      request.Filter.Id,
			Email:   request.Filter.Email,
			OrgID:   request.Filter.OrgId,
			SpaceID: request.Filter.SpaceId,
			OwnerID: request.Filter.OwnerId,
			Role:    request.Filter.Role,
		}
	}

	if request.Options != nil {
		opts = options.FindOptionsFromPB(request.Options)
	}

	return request
}

func (e *Encoder) DecodeFindRequest(request *pb.FindRequest) (filter *invitations.Filter, opts *options.FindOptions) {
	if request == nil {
		return
	}

	if request.Filter != nil {
		filter = &invitations.Filter{
			ID:      request.Filter.Id,
			Email:   request.Filter.Email,
			OrgID:   request.Filter.OrgId,
			SpaceID: request.Filter.SpaceId,
			OwnerID: request.Filter.OwnerId,
			Role:    request.Filter.Role,
		}
	}

	if request.Options != nil {
		opts = options.FindOptionsFromPB(request.Options)
	}
	return
}

func (e *Encoder) EncodeFindResponse(response *pb.FindResponse) (invitations []*invitations.Invitation, total int) {
	if response == nil {
		return
	}

	total = int(response.Total)
	for _, pi := range response.Invitations {
		invitations = append(invitations, InvitationFromProto(pi))
	}
	return
}

func (e *Encoder) DecodeFindResponse(invitations []*invitations.Invitation, total int) *pb.FindResponse {
	protoInvitations := make([]*pb.Invitation, 0, len(invitations))
	for _, i := range invitations {
		protoInvitations = append(protoInvitations, InvitationToProto(i))
	}
	return &pb.FindResponse{
		Invitations: protoInvitations,
		Total:       int64(total),
	}
}

func (e *Encoder) EncodeGetRequest(invitationId string) *pb.GetRequest {
	return &pb.GetRequest{
		InvitationId: invitationId,
	}
}

func (e *Encoder) DecodeGetRequest(request *pb.GetRequest) (invitationId string) {
	if request == nil {
		return
	}
	return request.InvitationId
}

func (e *Encoder) EncodeGetResponse(response *pb.GetResponse) (invitation *invitations.Invitation) {
	if response == nil {
		return
	}
	return InvitationFromProto(response.Invitation)
}

func (e *Encoder) DecodeGetResponse(invitation *invitations.Invitation) *pb.GetResponse {
	return &pb.GetResponse{
		Invitation: InvitationToProto(invitation),
	}
}

func InvitationToProto(invitation *invitations.Invitation) *pb.Invitation {
	if invitation == nil {
		return nil
	}
	pi := &pb.Invitation{
		Id:      invitation.ID,
		Email:   invitation.Email,
		OrgId:   invitation.OrgID,
		SpaceId: invitation.SpaceID,
		OwnerId: invitation.OwnerID,
		Role:    invitation.Role,
	}
	if invitation.CreatedAt != nil && !invitation.CreatedAt.IsZero() {
		pi.CreatedAt = timestamp.New(*invitation.CreatedAt)
	}
	if invitation.ValidUntil != nil && !invitation.ValidUntil.IsZero() {
		pi.ValidUntil = timestamp.New(*invitation.ValidUntil)
	}
	return pi
}

func InvitationFromProto(protoInvitation *pb.Invitation) *invitations.Invitation {
	if protoInvitation == nil {
		return nil
	}
	i := &invitations.Invitation{
		ID:      protoInvitation.Id,
		Email:   protoInvitation.Email,
		OrgID:   protoInvitation.OrgId,
		SpaceID: protoInvitation.SpaceId,
		OwnerID: protoInvitation.OwnerId,
		Role:    protoInvitation.Role,
	}
	if protoInvitation.CreatedAt != nil {
		t := protoInvitation.CreatedAt.AsTime()
		i.CreatedAt = &t
	}
	if protoInvitation.ValidUntil != nil {
		t := protoInvitation.ValidUntil.AsTime()
		i.ValidUntil = &t
	}
	return i
}
